#include <SDL/SDL_image.h>
#include "background.h"
#include "common.h"
#include "config.h"
#include "debug.h"
#include "files.h"
#include "options.h"

/*
 * Global variables
 */
SDL_Surface *background_picture, *logo;

/*
 * init_background	- loads all images that functions in this file needs
 *
 * IN:
 *  no parameters
 * RETURN VALUE:
 *  1 ...   on success
 *  0 ...   on error
 */
int
init_background(void)
{
    background_picture = try_load_image(BACKGROUND_FILE, -1, 0);
    if (!background_picture)
	return 0;
    logo = try_load_image(LOGO_FILE_NAME, -1, 1);
    if (!logo) {
	SDL_FreeSurface(background_picture);
	return 0;
    }
    return 1;
}

/*
 * free_background	- cleans memory allocated by init_background
 *
 * IN:
 *  no parameters
 * RETURN VALUE:
 *  1 ...   on success
 *  0 ...   in case that free_background was called before
 */
int
free_background(void)
{
    if (!background_picture)
	return 0;
    SDL_FreeSurface(background_picture);
    SDL_FreeSurface(logo);
    return 1;
}

/*
 * draw_background - draws part or whole background on the surface
 *
 * IN:
 *  to		use this surface
 *  dest	draw this area of surface or whole surface when NULL
 * RETURN VALUE:
 *  none
 */
void
draw_background2(SDL_Surface *to, SDL_Rect *dest)
{
    Sint16 x1, y1, x2, y2, x1_tile_start, y1_tile_start, x2_tile_start, y2_tile_start, x_counter, y_counter, width, height;
    SDL_Rect mydest;

    if (dest) {
	x1 = dest->x;
	y1 = dest->y;
	x1_tile_start = (x1 / background_picture->w + 1) * background_picture->w;
	y1_tile_start = (y1 / background_picture->h + 1) * background_picture->h;
	x2 = x1 + dest->w;
	y2 = y1 + dest->h;
	draw_background_tile(to, x1, y1, dest->w, dest->h);
	for (y_counter = y1_tile_start; y_counter < y2; y_counter += background_picture->h)
	    draw_background_tile(to, x1, y_counter, dest->w, y2 - y_counter);
	for (x_counter = x1_tile_start; x_counter < x2; x_counter += background_picture->w)
	    draw_background_tile(to, x_counter, y1, x2 - x_counter, dest->h);
    } else {
	x1 = 0;
	y1 = 0;
	x1_tile_start = 0;
	y1_tile_start = 0;
	x2 = to->w;
	y2 = to->h;
    }
    x2_tile_start = x2 - (x2 % background_picture->w);
    y2_tile_start = y2 - (y2 % background_picture->h);
    mydest.w = background_picture->w;
    mydest.h = background_picture->h;
    for (y_counter = y1_tile_start; y_counter < y2_tile_start; y_counter += background_picture->h) {
	mydest.y = y_counter;
	for (x_counter = x1_tile_start; x_counter < x2_tile_start; x_counter += background_picture->w) {
	    mydest.x = x_counter;
	    DRAW_BACKGROUND_TILE2(to, mydest);
	}
    }
    width = x2 - x2_tile_start;
    height = y2 - y2_tile_start;
    if (width && height && x2_tile_start >= x1 && y2_tile_start >= y1)
	draw_background_tile(to, x2_tile_start, y2_tile_start, width, height);
    for (y_counter = y1_tile_start; y_counter < y2_tile_start; y_counter += background_picture->h)
	    draw_background_tile(to, x2_tile_start, y_counter, width, background_picture->h);
    for (x_counter = x1_tile_start; x_counter < x2_tile_start; x_counter += background_picture->w)
	    draw_background_tile(to, x_counter, y2_tile_start, background_picture->h, height);
}

/*
 * draw_background_tile	- draws part of one background tile image
 *
 * IN:
 *  to		use this surface
 *  x		where to start in horizontal direction
 *  y		where to start in vertical direction
 *  w		width of part (may be greater than tile)
 *  h		height of part (may be greater than tile)
 * RETURN VALUE:
 *  none
 */
void
draw_background_tile(SDL_Surface *to, const Sint16 x, const Sint16 y, const Uint16 w, const Uint16 h)
{
    SDL_Rect src, dest;

    dest.x = x;
    dest.y = y;
    src.x = x % background_picture->w;
    src.y = y % background_picture->h;
    src.w = w;
    src.h = h;
    if (src.x + src.w > background_picture->w)
	src.w = background_picture->w - src.x;
    if (src.y + src.h > background_picture->h)
	src.h = background_picture->h - src.y;
    dest.w = src.w;
    dest.h = src.h;
    SDL_BlitSurface(background_picture, &src, to, &dest);
}

/*
 * draw_background - draws part or whole background on the surface
 *
 * IN:
 *  to		use this surface
 *  dest	draw this area of surface or whole surface when NULL
 * RETURN VALUE:
 *  none
 */
void
draw_background(SDL_Surface *to, SDL_Rect *dest)
{
    Sint16 x1, y1, x2, y2, y_cur;
    Uint16 w1, w2 = 0, h1, h2 = 0, w, h;

    if (dest) {
	x1 = dest->x;
	y1 = dest->y;
	x2 = dest->x + dest->w;
	if (x2 > to->w)
	    x2 = to->w;
	y2 = dest->y + dest->h;
	if (y2 > to->h)
	    y2 = to->h;
    } else {
	x1 = 0;
	y1 = 0;
	x2 = to->w;
	y2 = to->h;
    }
    w = x2 - x1;
    h = y2 - y1;
    w1 = background_picture->w - x1 % background_picture->w;
    h1 = background_picture->h - y1 % background_picture->h;
    if (w1 > w)
	w1 = w;
    else {
	if (w1 != w)
	    w2 = x2 % background_picture->w;
    }
    if (h1 > h)
	h1 = h;
    else {
	if (h1 != h)
	    h2 = y2 % background_picture->h;
    }
    y_cur = y1;
    draw_background_horiz(to, x1, y_cur, x2, w1, w2, h1);
    y_cur += h1;
    while (y_cur + background_picture->h <= y2) {
	draw_background_horiz(to, x1, y_cur, x2, w1, w2, background_picture->h);
	y_cur += background_picture->h;
    }
    if (h2 > 0)
	draw_background_horiz(to, x1, y_cur, x2, w1, w2, h2);
}

/*
 * draw_background_horiz	- draws horizontal area of tiles
 *
 * IN:
 *  to		use this surface
 *  x1		where to start in horizontal direction
 *  y1		where to start in vertical direction
 *  x2		where to end in horizontal direction
 *  w1		width of first part
 *  w2		width of last part
 *  h		height of part (may be greater than tile)
 * RETURN VALUE:
 *  none
 */
void
draw_background_horiz(SDL_Surface *to, const Sint16 x1, const Sint16 y1, const Sint16 x2, const Uint16 w1, const Uint16 w2, const Uint16 h)
{
    SDL_Rect src, dest;

    src.x = x1 % background_picture->w;
    src.y = y1 % background_picture->h;
    src.w = dest.w = w1;
    src.h = dest.h = h;
    dest.x = x1;
    dest.y = y1;
    SDL_BlitSurface(background_picture, &src, to, &dest);
    dest.x += dest.w;
    src.x = 0;
    src.w = dest.w = background_picture->w;
    while (dest.x + dest.w <= x2) {
	SDL_BlitSurface(background_picture, &src, to, &dest);
	dest.x += dest.w;
    }
    if (w2 > 0) {
	dest.w = src.w = w2;
	SDL_BlitSurface(background_picture, &src, to, &dest);
    }
}

/*
 * draw_logo	- draws part or whole logo image in the center of the screen
 *
 * IN:
 *  to		use this surface
 *  dest	if NULL then draw whole logo, otherwise only it's part
 * RETURN VALUE:
 *  none
 */
void
draw_logo(SDL_Surface *to, SDL_Rect *dest)
{
    SDL_Rect src, new_dest, logo_rect;

    logo_rect.w = logo->w;
    logo_rect.h = logo->h;
    logo_rect.x = (screen_width - logo_rect.w) / 2;
    logo_rect.y = (screen_height - logo_rect.h) / 2;
    if (dest)
	memcpy((void *)&new_dest, (void *)dest, sizeof(SDL_Rect));
    else
	memcpy((void *)&new_dest, (void *)&logo_rect, sizeof(SDL_Rect));
    if (new_dest.x < logo_rect.x) {
	new_dest.w -= logo_rect.x - new_dest.x;
	new_dest.x = logo_rect.x;
    }
    if (new_dest.y < logo_rect.y) {
	new_dest.h -= logo_rect.y - new_dest.y;
	new_dest.y = logo_rect.y;
    }
    if (new_dest.w <= 0 || new_dest.h <= 0)
	return;
    src.w = new_dest.w;
    src.h = new_dest.h;
    src.x = new_dest.x - logo_rect.x;
    src.y = new_dest.y - logo_rect.y;
    SDL_BlitSurface(logo, &src, to, &new_dest);
}
