#include <assert.h>
#include <stdlib.h>
#include "explosion.h"
#include "common.h"
#include "files.h"
#include "debug.h"
#if DEBUG > NO_DEBUG
#   include <stdio.h>
#endif /* DEBUG */

extern Sint16 screen_width, screen_height;

SDL_Surface *explosion = NULL;
explosion_type *explosions_array = NULL;
int explosions_count = 0, frames_count, frame_height;

void
init_explosions(void)
{
    assert(explosion == NULL);
    explosion = try_load_image(EXPLOSION_FILE_NAME, -1, 1);
#ifdef EXPLOSION_FRAMES_COUNT
    frames_count = EXPLOSION_FRAMES_COUNT;
#else /* !EXPLOSION_FRAMES_COUNT */
    frames_count = explosion->h / explosion->w;
#endif /* !EXPLOSION_FRAMES_COUNT */
    frame_height = explosion->h / frames_count;
#if DEBUG >= EXTRA_DEBUG
    printf("Explosion module initialised.\n");
#endif /* DEBUG */
}

void
free_explosions(void)
{
    int counter;
    
    SDL_FreeSurface(explosion);
    for (counter = 0; counter < explosions_count; counter++)
	free_explosion(counter);
    if (explosions_array)
	free(explosions_array);
    explosions_count = 0;
    explosions_array = NULL;
    explosion = NULL;
}

void
free_explosion(const int explosion_number)
{
    /* Nothing to do */
}

void
create_explosion(object_type *owner)
{
    assert(explosion);
    if (explosions_count++ > 0) {
	explosions_array = (explosion_type *)realloc((void *)explosions_array, sizeof(explosion_type)*explosions_count);
    }
    else {
	explosions_array = (explosion_type *)malloc(sizeof(explosion_type));
    }
    init_space_object((space_object_type *)&explosions_array[explosions_count - 1]);
    explosions_array[explosions_count - 1].type = EXPLOSION_TYPE;
    explosions_array[explosions_count - 1].explosion_image = explosion;
    explosions_array[explosions_count - 1].explosion_position_left = owner->space_object.object_position_left;
    explosions_array[explosions_count - 1].explosion_position_top = owner->space_object.object_position_top;
    explosions_array[explosions_count - 1].current_frame = -1;
}

void
delete_all_explosions(SDL_Surface *to)
{
    int counter;

    for (counter = 0; counter < explosions_count; counter++) {
	clean_explosion(counter, to);
	purge_explosion(counter, to);
	free_explosion(counter);
    }
    if (explosions_array)
	free(explosions_array);
    explosions_count = 0;
    explosions_array = NULL;
}

void
draw_explosion(const int explosion_number, SDL_Surface *to)
{
    SDL_Rect imgrect;

    assert(explosion_number < explosions_count);
    if (explosions_array[explosion_number].is_shown)
	return;
    imgrect.x = 0;
    imgrect.y = frame_height * explosions_array[explosion_number].current_frame;
    imgrect.w = explosion->w;
    imgrect.h = frame_height;
    explosions_array[explosion_number].update_rect.x = explosions_array[explosion_number].explosion_position_left - (imgrect.w / 2);
    explosions_array[explosion_number].update_rect.y = explosions_array[explosion_number].explosion_position_top - (imgrect.w / 2);
    explosions_array[explosion_number].update_rect.w = imgrect.w;
    explosions_array[explosion_number].update_rect.h = imgrect.h;
    if (explosions_array[explosion_number].update_rect.x < 0) {
	imgrect.w += explosions_array[explosion_number].update_rect.x;
	imgrect.x -= explosions_array[explosion_number].update_rect.x;
	explosions_array[explosion_number].update_rect.x = 0;
    }
    if (explosions_array[explosion_number].update_rect.y < 0) {
	imgrect.h += explosions_array[explosion_number].update_rect.y;
	imgrect.y -= explosions_array[explosion_number].update_rect.y;
	explosions_array[explosion_number].update_rect.y = 0;
    }
    if (explosions_array[explosion_number].update_rect.x + imgrect.w >= screen_width)
	imgrect.w = screen_width - explosions_array[explosion_number].update_rect.x;
    if (explosions_array[explosion_number].update_rect.y + imgrect.h >= screen_height)
	imgrect.h = screen_height - explosions_array[explosion_number].update_rect.y;
    explosions_array[explosion_number].update_rect.w = imgrect.w;
    explosions_array[explosion_number].update_rect.h = imgrect.h;
    SDL_BlitSurface(explosion, &imgrect, to, &explosions_array[explosion_number].update_rect);
    explosions_array[explosion_number].need_update |= DRAW_UPDATE;
    explosions_array[explosion_number].is_shown = 1;
}

void
update_explosion(const int explosion_number, SDL_Surface *to)
{
    assert(explosion_number < explosions_count);
    update_space_object((space_object_type *)&explosions_array[explosion_number], to);
}

void
clean_explosion(const int explosion_number, SDL_Surface *to)
{
    assert(explosion_number < explosions_count);
    clean_space_object((space_object_type *)&explosions_array[explosion_number], to);
}

void
purge_explosion(const int explosion_number, SDL_Surface *to)
{
    assert(explosion_number < explosions_count);
    purge_space_object((space_object_type *)&explosions_array[explosion_number], to);
}

void
continue_exploding(SDL_Surface *to)
{
    int counter, all_done = 1;
    
    for (counter = 0; counter < explosions_count; counter++) {
	if (explosions_array[counter].current_frame <= frames_count)
	    explosions_array[counter].current_frame++;
	else
	    continue;
	all_done = 0;
	clean_explosion(counter, to);
	if (explosions_array[counter].current_frame == frames_count)
	    purge_explosion(counter, to);
	else
	    draw_explosion(counter, to);
    }
    for (counter = 0; counter < explosions_count; counter++)
	update_explosion(counter, to);
    if (all_done)
	delete_all_explosions(to);
}
