#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "config.h"
#include "debug.h"
#include "input.h"
#ifdef DEBUG
#   include <stdio.h>
#endif /* DEBUG */

typedef Uint8 actions_type[ACTIONS_COUNT];

char *input_method_description[] = {
    "no input", "keyboard", "joystick"
};

#ifdef ENABLE_JOYSTICK
#include "ask.h"
#include "background.h"
#include "common.h"

#define CAL_COMMANDS_COUNT  3

const char *calibrate_commands[CAL_COMMANDS_COUNT] = {
    "Center joystick and press any button",
    "Move joystick in all directions and press any button",
    "Test your joystick and press any button"
};

const Uint16 test_box_width = 100, test_box_height = 100, text_height = 100;
const int text_ptsize = 16;
#endif /* ENABLE_JOYSTICK */

actions_type *active_actions = NULL;
Uint8 system_actions = 0;
input_type *players_input = NULL;
int max_players_count;

void
init_input(const int max_players)
{
    max_players_count = max_players;
    active_actions = (actions_type *)malloc(sizeof(actions_type) * max_players);
    players_input = (input_type *)malloc(sizeof(input_type) * max_players);
    memset((void *)active_actions, 0, sizeof(actions_type) * max_players);
    memset((void *)players_input, 0, sizeof(input_type) * max_players);
    SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
    SDL_EventState(SDL_MOUSEBUTTONDOWN, SDL_IGNORE);
    SDL_EventState(SDL_MOUSEBUTTONUP, SDL_IGNORE);
    SDL_EventState(SDL_ACTIVEEVENT, SDL_IGNORE);
#ifdef ENABLE_JOYSTICK
    SDL_JoystickEventState(SDL_ENABLE);
    memset((void *)joysticks, 0, sizeof(SDL_Joystick *) * MAX_JOYS_COUNT);
    memset((void *)joysticks_params, 0, sizeof(joy_params_type) * MAX_JOYS_COUNT);
#endif /* ENABLE_JOYSTICK */
}

void
close_input(void)
{
#ifdef ENABLE_JOYSTICK
    int counter;
    
    for (counter = 0; counter < SDL_NumJoysticks(); counter++)
	if (SDL_JoystickOpened(counter))
	    SDL_JoystickClose(joysticks[counter]);
#endif /* ENABLE_JOYSTICK */
    free(active_actions);
    active_actions = NULL;
    free(players_input);
    players_input = NULL;
}

int
is_action_active(const int which_player, const player_actions action)
{
    assert(which_player < max_players_count);
    assert(active_actions);
    while (work_events());
    return (int)active_actions[which_player][ACTION_INDEX(action)];
}

int
get_system_requests(void)
{
    int result;

    while (work_events());
    result = (int)system_actions;
    system_actions = 0;
    return result;
}

int
set_action_key(const int which_player, const player_actions action, const SDLKey key)
{
    assert(players_input);
    assert(which_player < max_players_count);
    if (players_input->type != keyboard)
	return ERR_NO_RIGHT_INPUT_SELECTED;
    players_input[which_player].keyboard.keys[ACTION_INDEX(action)] = key;
    return NO_ERROR;
}

SDLKey
get_action_key(const int which_player, const player_actions action)
{
    assert(players_input);
    assert(which_player < max_players_count);
    if (players_input->type != keyboard)
	return 0;
    return players_input[which_player].keyboard.keys[ACTION_INDEX(action)];
}

#ifdef ENABLE_JOYSTICK
int
open_joystick(const Uint8 which)
{
    if (which >= MAX_JOYS_COUNT || which >= SDL_NumJoysticks())
	return ERR_BAD_JOYSTICK_NUMBER;
    if (SDL_JoystickOpened(which))
	return NO_ERROR;
    joysticks[which] = SDL_JoystickOpen(which);
    if (!joysticks[which]) {
	fprintf(stderr, "Couldn't open joystick %d: %s!\n", (int)which, SDL_GetError());
	return ERR_SDL_ERROR;
    }
    joysticks_params[which].x_left = AXIS_MIN_DEFAULTS;
    joysticks_params[which].x_right = AXIS_MAX_DEFAULTS;
    joysticks_params[which].y_up = AXIS_MIN_DEFAULTS;
    joysticks_params[which].y_down = AXIS_MAX_DEFAULTS;
    return NO_ERROR;
}

int
set_action_joy_button(const int which_player, const player_actions action, const Uint8 button)
{
    assert(players_input);
    assert(which_player < max_players_count);
    if (players_input[which_player].type != joystick || !is_joystick_active(which_player))
	return ERR_NO_RIGHT_INPUT_SELECTED;
    if (button >= SDL_JoystickNumButtons(joysticks[players_input[which_player].joystick.joy_number])) {
#if DEBUG >= EXTRA_DEBUG
	fprintf(stderr, "set_action_joy_button: bad button number - %d\n", (int)button);
#endif /* DEBUG */
	return ERR_BAD_BUTTON_NUMBER;
    }
    switch (action) {
	case fire:
	    players_input[which_player].joystick.fire_button = button;
	    break;
	case fire_blast:
	    players_input[which_player].joystick.blast_button = button;
	    break;
	default:
#if DEBUG >= EXTRA_DEBUG
	    fprintf(stderr, "set_action_joy_button: bad action number - %d\n", (int)action);
#endif /* DEBUG */
	    return ERR_NOT_IMPLEMENTED;
    }
    return NO_ERROR;
}

int
set_joystick(const int which_player, const Uint8 joystick_number)
{
    int result;
    
    assert(players_input);
    assert(which_player < max_players_count);
    if (players_input[which_player].type == joystick) {
	result = open_joystick(joystick_number);
	players_input[which_player].joystick.joy_active = 0;
	if (result == NO_ERROR) {
	    players_input[which_player].joystick.joy_number = joystick_number;
	    players_input[which_player].joystick.joy_active = 1;
	}
	else
	    players_input[which_player].joystick.joy_number = 0;
	return result;
    } else
	return NO_ERROR;
}
#endif /* ENABLE_JOYSTICK */

void
set_players_input_method(const int which_player, const input_method method)
{
    assert(players_input);
    assert(which_player < max_players_count);
    if (players_input[which_player].type == method)
	return;
    memset((void *)(players_input + which_player), 0, sizeof(input_type));
    players_input[which_player].type = method;
}

int
work_events(void)
{
    SDL_Event event;
    SDLKey key;
    int counter1, counter2;
#ifdef ENABLE_JOYSTICK
    Uint8 joy_num;
#endif /* ENABLE_JOYSTICK */

    if (SDL_PollEvent(&event)) {
	switch (event.type) {
	    case SDL_KEYDOWN:
	    case SDL_KEYUP:
		key = ((SDL_KeyboardEvent *)&event)->keysym.sym;
		if (key == QUIT_KEY && event.type == SDL_KEYDOWN) {
		    system_actions |= QUIT_EVENT;
		    return ONE_MORE;
		}
		for (counter1 = 0; counter1 < max_players_count; counter1++)
		    if (players_input[counter1].type == keyboard)
			for (counter2 = 0; counter2 < ACTIONS_COUNT; counter2++)
			    if (players_input[counter1].keyboard.keys[counter2] == key) {
				active_actions[counter1][counter2] = (event.type == SDL_KEYDOWN ? 1 : 0);
				return ONE_MORE;
			    }
		return ONE_MORE;
#ifdef ENABLE_JOYSTICK
	    case SDL_JOYBUTTONDOWN:
	    case SDL_JOYBUTTONUP:
		for (counter1 = 0; counter1 < max_players_count; counter1++)
		    if (players_input[counter1].type == joystick && (int)players_input[counter1].joystick.joy_number == event.jbutton.which) {
			if (players_input[counter1].joystick.fire_button == event.jbutton.button) {
				active_actions[counter1][ACTION_INDEX(fire)] = (event.type == SDL_JOYBUTTONDOWN ? 1 : 0);
				return ONE_MORE;
			}
			if (players_input[counter1].joystick.blast_button == event.jbutton.button)
				active_actions[counter1][ACTION_INDEX(fire_blast)] = (event.type == SDL_JOYBUTTONDOWN ? 1 : 0);
			return ONE_MORE;
		    }
		return ONE_MORE;
	    case SDL_JOYAXISMOTION:
		if (event.jaxis.axis > 1)
		    return ONE_MORE;
		for (counter1 = 0; counter1 < max_players_count; counter1++)
		    if (players_input[counter1].type == joystick && (int)get_joystick(counter1) == event.jbutton.which) {
			joy_num = get_joystick(counter1);
			if (event.jaxis.axis == 0) {
			    if (event.jaxis.value > joysticks_params[joy_num].x_right) {
				active_actions[counter1][ACTION_INDEX(rotate_right)] = 1;
				active_actions[counter1][ACTION_INDEX(rotate_left)] = 0;
			    } else {
				if (event.jaxis.value < joysticks_params[joy_num].x_left) {
				    active_actions[counter1][ACTION_INDEX(rotate_right)] = 0;
				    active_actions[counter1][ACTION_INDEX(rotate_left)] = 1;
				} else {
				    active_actions[counter1][ACTION_INDEX(rotate_right)] = 0;
				    active_actions[counter1][ACTION_INDEX(rotate_left)] = 0;
				}
			    }
			} else {
			    if (event.jaxis.value > joysticks_params[joy_num].y_down) {
				active_actions[counter1][ACTION_INDEX(speed_up)] = 0;
				active_actions[counter1][ACTION_INDEX(speed_down)] = 1;
			    } else {
				if (event.jaxis.value < joysticks_params[joy_num].y_up) {
				    active_actions[counter1][ACTION_INDEX(speed_up)] = 1;
				    active_actions[counter1][ACTION_INDEX(speed_down)] = 0;
				} else {
				    active_actions[counter1][ACTION_INDEX(speed_up)] = 0;
				    active_actions[counter1][ACTION_INDEX(speed_down)] = 0;
				}
			    }
			}
			return ONE_MORE;
		    }
		    return ONE_MORE;
#endif /* ENABLE_JOYSTICK */
	    case SDL_QUIT:
		system_actions |= QUIT_EVENT;
		return ONE_MORE;
	    default:
		return ONE_MORE;
	}
    }
    else
	return ALL_DONE;
}

SDLKey
get_key(void)
{
    SDL_Event event;
    do {
	if (!SDL_WaitEvent(&event))
	    return (SDLKey)0;
	if (event.type == SDL_KEYDOWN)
	    return ((SDL_KeyboardEvent *)&event)->keysym.sym;
	else
	    continue;
    } while (1);
    return (SDLKey)0;
}

input_method
get_players_input_method(const int which_player)
{
    assert(players_input);
    assert(which_player < max_players_count);
    return players_input[which_player].type;
}

char *
get_input_method_desc(const input_method method)
{
    return input_method_description[method];
}

#ifdef ENABLE_JOYSTICK
Uint8
get_action_joy_button(const int which_player, const player_actions action)
{
    assert(players_input);
    assert(which_player < max_players_count);
    if (players_input[which_player].type != joystick || !players_input[which_player].joystick.joy_active)
	return 255;
    switch (action) {
	case fire:
	    return players_input[which_player].joystick.fire_button;
	case fire_blast:
	    return players_input[which_player].joystick.blast_button;
	default:
#if DEBUG >= EXTRA_DEBUG
	    fprintf(stderr, "get_action_joy_button: bad action number - %d\n", (int)action);
#endif /* DEBUG */
	    return 255;
    }
}

int
is_joystick_active(const int which_player)
{
    assert(players_input);
    assert(which_player < max_players_count);
    return players_input[which_player].joystick.joy_active;
}

Uint8
get_joystick(const int which_player)
{
    assert(players_input);
    assert(which_player < max_players_count);
    if (players_input[which_player].type != joystick || !players_input[which_player].joystick.joy_active)
	return 0;
    return players_input[which_player].joystick.joy_number;
}

void
calibrate_joystick(SDL_Surface *to, const int which_player)
{
    int counter;
    Sint16 x_center = 0, y_center = 0, x_max = 0, y_max = 0, x_min = 0, y_min = 0, cur_x, cur_y;
    SDL_Rect dst = { 0, 0, 0, 0 }, rect, rect2;
    SDL_Color color = { 0xFF, 0xFF, 0xFF, 0 }, bg = { 0x0, 0x0, 0x0, 0 };
    SDL_Event recv_event;
    SDL_Joystick *this_joystick;

    if (get_players_input_method(which_player) != joystick || (int)get_joystick(which_player) >= SDL_NumJoysticks() || !players_input[which_player].joystick.joy_active)
	return;
    draw_background(to, NULL);
    draw_logo(to, NULL);
    SDL_Flip(to);
    dst.w = to->w;
    dst.h = to->h;
    rect.w = test_box_width + 2;
    rect.h = test_box_height + 2;
    rect.x = (to->w - rect.w) / 2;
    rect.y = (to->h - rect.h) / 2;
    draw_rect(to, &rect, &color);
    SDL_UpdateRects(to, 1, &rect);
    if (open_joystick(get_joystick(which_player)) != NO_ERROR)
	return;
    this_joystick = joysticks[get_joystick(which_player)];
    for (counter = 0; counter < CAL_COMMANDS_COUNT; counter++) {
	if (counter) {
	    draw_background(to, &dst);
	    draw_logo(to, &dst);
	    SDL_UpdateRects(to, 1, &dst);
	}
	draw_text_centered(to, calibrate_commands[counter], text_ptsize, text_height, color, bg, &dst);
	switch (counter) {
	    case 0:
		for (;;) {
		    SDL_WaitEvent(&recv_event);
		    if (recv_event.type == SDL_JOYBUTTONDOWN && recv_event.jbutton.which == get_joystick(which_player)) {
			SDL_JoystickUpdate();
			x_center = SDL_JoystickGetAxis(this_joystick, 0);
			y_center = SDL_JoystickGetAxis(this_joystick, 1);
			x_min = x_max = x_center;
			y_min = y_max = y_center;
			break;
		    } else
			if ((recv_event.type == SDL_KEYDOWN && recv_event.key.keysym.sym == QUIT_EVENT) ||
				(recv_event.type == SDL_QUIT)) {
			    SDL_PushEvent(&recv_event);
			    return;
			}
		}
		break;
	    case 1:
		for (;;) {
		    SDL_WaitEvent(&recv_event);
		    if (recv_event.type == SDL_JOYBUTTONDOWN && recv_event.jbutton.which == get_joystick(which_player))
			break;
		    else {
			if ((recv_event.type == SDL_KEYDOWN && recv_event.key.keysym.sym == QUIT_EVENT) ||
				(recv_event.type == SDL_QUIT)) {
			    SDL_PushEvent(&recv_event);
			    return;
			} else
			    if (recv_event.type == SDL_JOYAXISMOTION && recv_event.jaxis.which == get_joystick(which_player)) {
				if (recv_event.jaxis.axis == 0) {
				    if (recv_event.jaxis.value > x_max)
					x_max = recv_event.jaxis.value;
				    if (recv_event.jaxis.value < x_min)
					x_min = recv_event.jaxis.value;
				}
				if (recv_event.jaxis.axis == 1) {
				    if (recv_event.jaxis.value > y_max)
					y_max = recv_event.jaxis.value;
				    if (recv_event.jaxis.value < y_min)
					y_min = recv_event.jaxis.value;
				}
			    }
		    }
		}
		break;
	    case 2:
		cur_x = test_box_width / 2;
		cur_y = test_box_height / 2;
		rect.x++;
		rect.y++;
		rect.w -= 2;
		rect.h -= 2;
		SDL_SetClipRect(to, &rect);
		draw_cross(to, cur_x + rect.x, cur_y + rect.y, 11, &color);
		rect2.w = 11 + 1;
		rect2.h = 11 + 1;
		rect2.x = cur_x + rect.x - 11 / 2;
		rect2.y = cur_y + rect.y - 11 / 2;
		SDL_UpdateRects(to, 1, &rect2);
		for (;;) {
		    SDL_WaitEvent(&recv_event);
		    if (recv_event.type == SDL_JOYBUTTONDOWN && recv_event.jbutton.which == get_joystick(which_player))
			break;
		    else {
			if ((recv_event.type == SDL_KEYDOWN && recv_event.key.keysym.sym == QUIT_EVENT) ||
				(recv_event.type == SDL_QUIT)) {
			    SDL_SetClipRect(to, NULL);
			    SDL_PushEvent(&recv_event);
			    return;
			} else
			    if (recv_event.type == SDL_JOYAXISMOTION && recv_event.jaxis.which == get_joystick(which_player)) {
				draw_background(to, &rect2);
				draw_logo(to, &rect2);
				SDL_UpdateRects(to, 1, &rect2);
				if (recv_event.jaxis.axis == 0) {
				    if (recv_event.jaxis.value > x_center)
					cur_x = test_box_width / 2 + ((test_box_width / 2) * (recv_event.jaxis.value - x_center)) / (x_max - x_center);
				    else
					cur_x = ((test_box_width / 2) * recv_event.jaxis.value) / (x_center - x_min);
				}
				if (recv_event.jaxis.axis == 1) {
				    if (recv_event.jaxis.value > y_center)
					cur_y = test_box_height / 2 + ((test_box_width / 2) * (recv_event.jaxis.value - y_center)) / (y_max - y_center);
				    else
					cur_y = ((test_box_height / 2) * recv_event.jaxis.value) / (y_center - y_min);
				}
				draw_cross(to, cur_x + rect.x, cur_y + rect.y, 11, &color);
				rect2.x = cur_x + rect.x - 11 / 2;
				rect2.y = cur_y + rect.y - 11 / 2;
				rect2.w = 11 + 1;
				rect2.h = 11 + 1;
				if (rect2.x < rect.x)
				    rect2.x = rect.x;
				if (rect2.y < rect.y)
				    rect2.y = rect.y;
				if (rect2.x + rect2.w > rect.x + rect.w)
				    rect2.w = rect.w + rect.x - rect2.x;
				if (rect2.y + rect2.h > rect.y + rect.h)
				    rect2.h = rect.h + rect.y - rect2.y;
				SDL_UpdateRects(to, 1, &rect2);
			    }
		    }
		}
		SDL_SetClipRect(to, NULL);
		break;
	}
    }
    joysticks_params[get_joystick(which_player)].x_left = x_center - (x_center - x_min) / 3;
    joysticks_params[get_joystick(which_player)].x_right = x_center + (x_max - x_center) / 3;
    joysticks_params[get_joystick(which_player)].y_up = y_center - (y_center - y_min) / 3;
    joysticks_params[get_joystick(which_player)].y_down = y_center + (y_max - y_center) / 3;
}

Uint8
get_button(const int which_player)
{
    SDL_Event recv_event;
    SDL_Joystick *this_joystick;

    if (get_players_input_method(which_player) != joystick || (int)get_joystick(which_player) >= SDL_NumJoysticks() || !players_input[which_player].joystick.joy_active)
	return 255; 
    if (open_joystick(get_joystick(which_player)) != NO_ERROR)
	return 255;
    this_joystick = joysticks[get_joystick(which_player)];
    for (;;) {
	SDL_WaitEvent(&recv_event);
	if (recv_event.type == SDL_JOYBUTTONDOWN && recv_event.jbutton.which == get_joystick(which_player))
		return recv_event.jbutton.button;
	else
	    if ((recv_event.type == SDL_KEYDOWN && recv_event.key.keysym.sym == QUIT_EVENT) ||
		    (recv_event.type == SDL_QUIT)) {
		SDL_PushEvent(&recv_event);
		return 255;
	    }
    }
    return 255;
}
#endif /* ENABLE_JOYSTICK */
