#include <assert.h>
#include <stdlib.h>
#include "config.h"
#include "list.h"

/*
 * init_list	- inits list
 *
 * IN:
 *  no parameters
 * RETURN VALUE:
 *  head of new list
 */
struct node_type *
init_list()
{
    struct node_type *head, *tail;

    head = (struct node_type *)malloc(sizeof(struct node_type));
    tail = (struct node_type *)malloc(sizeof(struct node_type));
    head->prev = NULL;
    head->next = tail;
    head->object_ptr = NULL;
    tail->prev = head;
    tail->next = NULL;
    tail->object_ptr = NULL;

    return head;
}

/*
 * free_list	- destroys list
 *
 * IN:
 *  no parameters
 * RETURN VALUE:
 *  none
 */
void
free_list(struct node_type *root)
{
    assert(root);
    assert(!root->prev);
    clean_list(root);
    assert(root->next);
    free(root->next);
    free(root);
}

/*
 * clean_list	- destroys content of the list
 *
 * IN:
 *  root	pointer to list's head
 * RETURN VALUE:
 *  none
 */
void
clean_list(struct node_type *root)
{
    struct node_type *ptr, *tofree;

    assert(root);
    assert(!root->prev);
    assert(root->next);
    
    ptr = root->next;
    while (ptr->next) {
	tofree = ptr;
	ptr = ptr->next;
	free(tofree);
    }
    root->next = ptr;
    ptr->prev = root;
}

/*
 * insert_into_list	- inserts object into the sorted list
 *
 * IN:
 *  root	pointer to list's head
 *  object	pointer to object
 * RETURN VALUE:
 *  NULL	if root == NULL || object == NULL or other ERROR
 *  pointer to new node on success
 */
struct node_type *
insert_into_list(struct node_type *root, object_type *object)
{
    struct node_type *ptr, *new_node;
    
    if (!root || !object)
	return NULL;
    assert(root->next);
    assert(!root->prev);
    
    ptr = root->next;
    while (object > ptr->object_ptr && ptr->next)
	ptr = ptr->next;
    new_node = (struct node_type *)malloc(sizeof(struct node_type));
    new_node->object_ptr = object;
    new_node->next = ptr;
    new_node->prev = ptr->prev;
    ptr->prev->next = new_node;
    ptr->prev = new_node;

    return new_node;
}

/*
 * delete_from_list	- delete object from the sorted list
 *
 * IN:
 *  root	pointer to list's head
 *  object	pointer to object
 * RETURN VALUE:
 *  0		if root == NULL || object == NULL or object not found in list
 *  1		on success
 */
int
delete_from_list(struct node_type *root, object_type *object)
{
    struct node_type *tofree;
    
    if (!root || !object)
	return 0;
    assert(root->next);
    assert(!root->prev);
    tofree = search_in_list(root, object);
    if (!tofree)
	return 0;
    tofree->prev->next = tofree->next;
    tofree->next->prev = tofree->prev;
    free(tofree);

    return 1;
}

/*
 * search_in_list	- search object in the sorted list
 *
 * IN:
 *  root	pointer to list's head
 *  object	pointer to object
 * RETURN VALUE:
 *  NULL	if root == NULL || object == NULL or object not found in list
 *  pointer to node on success
 */
struct node_type *
search_in_list(struct node_type *root, object_type *object)
{
    struct node_type *ptr;
    
    if (!root || !object)
	return NULL;
    assert(root->next);
    assert(!root->prev);
    ptr = root->next;
    while (object > ptr->object_ptr && ptr->next)
	ptr = ptr->next;
    if (ptr->object_ptr == object)
	return ptr;
    else
	return NULL;
}

/*
 * get_first_object	- get first object in the list and initialize walk_ptr
 *
 * IN:
 *  root	pointer to list's head
 *  walk_ptr	address of pointer where is our current position
 * RETURN VALUE:
 *  NULL	if root == NULL || walk_ptr == NULL or there is no object in the list
 *  pointer to object on success
 */
object_type *
get_first_object(struct node_type *root, struct node_type **walk_ptr)
{
    if (!root || !walk_ptr)
	return NULL;
    assert(root->next);
    assert(!root->prev);
    if (!root->next->next)
	return NULL;
    *walk_ptr = root->next->next;

    return root->next->object_ptr;
}

/*
 * get_next_object	- get next object in the list and increase walk_ptr
 *
 * IN:
 *  walk_ptr	address of pointer where is our current position
 * RETURN VALUE:
 *  NULL	if walk_ptr == NULL or this is tail of the list
 *  pointer to object on success
 */
object_type *
get_next_object(struct node_type **walk_ptr)
{
    if (!walk_ptr || !*walk_ptr || !(*walk_ptr)->next)
	return NULL;
    *walk_ptr = (*walk_ptr)->next;

    return (*walk_ptr)->prev->object_ptr;
}
