#include <assert.h>
#ifdef DEBUG
#   include <stdio.h>
#endif /* DEBUG */
#include <string.h>
#include <math.h>
#include "background.h"
#include "objects.h"
#include "config.h"
#include "debug.h"

extern Sint16 screen_width, screen_height;

void
init_space_object(space_object_type *object)
{
    assert(object);
    memset((void *)object, 0, sizeof(space_object_type));
    object->images_count = 1;
}

void
free_space_object(space_object_type *object)
{
    assert(object);
    SDL_FreeSurface(object->object_image);
}

void
draw_space_object(space_object_type *object, SDL_Surface *to)
{
    SDL_Rect imgrect;

    assert(object);
    if (object->is_shown || (object->options & DO_NOT_DRAW))
	return;
    imgrect.x = 0;
    imgrect.y = 0;
    imgrect.w = object->object_image->w;
    imgrect.h = object->object_image->h;
    object->update_rect.x = object->object_position_left - (imgrect.w / 2);
    object->update_rect.y = object->object_position_top - (imgrect.w / 2);
    object->update_rect.w = imgrect.w;
    object->update_rect.h = imgrect.h;
    if (object->update_rect.x < 0) {
	imgrect.w += object->update_rect.x;
	imgrect.x -= object->update_rect.x;
	object->update_rect.x = 0;
    }
    if (object->update_rect.y < 0) {
	imgrect.h += object->update_rect.y;
	imgrect.y -= object->update_rect.y;
	object->update_rect.y = 0;
    }
    if (object->update_rect.x + imgrect.w >= screen_width)
	imgrect.w = screen_width - object->update_rect.x;
    if (object->update_rect.y + imgrect.h >= screen_height)
	imgrect.h = screen_height - object->update_rect.y;
    object->update_rect.w = imgrect.w;
    object->update_rect.h = imgrect.h;
    SDL_BlitSurface(object->object_image, &imgrect, to, &object->update_rect);
    object->need_update |= DRAW_UPDATE;
    object->is_shown = 1;
}

void
clean_space_object(space_object_type *object, SDL_Surface *to)
{
    Uint16 imgheight, imgwidth;

    assert(object);
    if (!object->is_shown)
	return;
    imgwidth = object->object_image->w;
    imgheight = object->object_image->h / object->images_count;
    object->clean_rect.w = imgwidth;
    object->clean_rect.h = imgheight;
    object->clean_rect.x = object->object_position_left - (imgwidth / 2);
    object->clean_rect.y = object->object_position_top - (imgheight / 2);
    if (object->clean_rect.x < 0) {
	object->clean_rect.w += object->clean_rect.x;
	object->clean_rect.x = 0;
    }
    if (object->clean_rect.y < 0) {
	object->clean_rect.h += object->clean_rect.y;
	object->clean_rect.y = 0;
    }
    if (object->clean_rect.x + object->clean_rect.w > screen_width)
	object->clean_rect.w = screen_width - object->clean_rect.x;
    if (object->clean_rect.y + object->clean_rect.h > screen_height)
	object->clean_rect.h = screen_height - object->clean_rect.y;
    draw_background(to, &object->clean_rect);
    object->need_update |= CLEAN_UPDATE;
    object->is_shown = 0;
}

void
update_space_object(space_object_type *object, SDL_Surface *to)
{
    assert(object);
    if (object->need_update & DRAW_UPDATE) {
	SDL_UpdateRects(to, 1, &object->update_rect);
	object->need_update &= ~DRAW_UPDATE;
    }
}

void
purge_space_object(space_object_type *object, SDL_Surface *to)
{
    assert(object);
    if (object->need_update & CLEAN_UPDATE) {
	SDL_UpdateRects(to, 1, &object->clean_rect);
	object->need_update &= ~CLEAN_UPDATE;
    }
}

void
do_move_with_space_object(space_object_type *object, SDL_Surface *to, enum wall_mode_type wall_mode)
{
    float angle, delta_fleft, delta_ftop;
    Sint16 delta_left, delta_top;

    assert(object);
    angle = object->object_angle;
    if (object->object_speed == 0)
	return;
    if (angle == 0 || angle == 180) {
	delta_ftop = object->object_distance_top;
	delta_fleft = (angle == 180 ? -1 : 1) * (object->object_speed / SPEED_CONST) + object->object_distance_left;
    }
    else
	if (angle == 90 || angle == 270) {
	    delta_ftop = (angle == 90 ? -1 : 1) * (object->object_speed / SPEED_CONST) + object->object_distance_top;
	    delta_fleft = object->object_distance_left;
	}
	else {
		angle *= M_PI / (float)180;
		delta_fleft = (object->object_speed / SPEED_CONST) * cos(angle) + object->object_distance_left;
		delta_ftop = - (object->object_speed / SPEED_CONST) * sin(angle) + object->object_distance_top;
	}
    delta_left = delta_fleft;
    delta_top = delta_ftop;
    object->object_distance_left = delta_fleft - delta_left;
    object->object_distance_top = delta_ftop - delta_top;
    change_space_object_position(object, delta_left, delta_top, to, wall_mode);
}

void
change_space_object_position(space_object_type *object, const Sint16 delta_left, const Sint16 delta_top, SDL_Surface *to, enum wall_mode_type wall_mode)
{
    Sint16 left, top;

    clean_space_object(object, to);
    left = object->object_position_left;
    top = object->object_position_top;
    left += delta_left;
    top += delta_top;
    if (left < 0) {
	switch (wall_mode) {
	    case stop_object:
		left = 0;
		object->object_speed = 0;
		break;
	    case go_through_wall:
		left = screen_width + left;
		break;
	    default:
		left -= delta_left;
		object->object_angle = 180 - object->object_angle;
	}
    }
    else
	if (left > screen_width) {
	    switch (wall_mode) {
		case stop_object:
		    left = screen_width;
		    object->object_speed = 0;
		    break;
		case go_through_wall:
		    left -= screen_width;
		    break;
		default:
		    left -= delta_left;
		    object->object_angle = 180 - object->object_angle;
	    }
	}
    if (top < 0) {
    	switch (wall_mode) {
	    case stop_object:
		top = 0;
		object->object_speed = 0;
		break;
	    case go_through_wall:
		top = screen_height + top;
		break;
	    default:
		top -= delta_top;
		object->object_angle = 360 - object->object_angle;
	}
    }
    else
	if (top > screen_height) {
	    switch (wall_mode) {
		case stop_object:
		    top = screen_height;
		    object->object_speed = 0;
		    break;
		case go_through_wall:
		    top -= screen_height;
		    break;
		default:
		    top -= delta_top;
		    object->object_angle = 360 - object->object_angle;
	    }
	}
    while (object->object_angle >= 360)
	object->object_angle -= 360;
    while (object->object_angle < 0)
	object->object_angle += 360;
    object->object_position_left = left;
    object->object_position_top = top;
    draw_space_object(object, to);
}
